package directrequest

import (
	"context"
	"fmt"
	"reflect"
	"sync"

	"github.com/smartcontractkit/chainlink/core/services/postgres"

	"github.com/ethereum/go-ethereum/common"
	"github.com/pkg/errors"
	"github.com/smartcontractkit/chainlink/core/assets"
	"github.com/smartcontractkit/chainlink/core/internal/gethwrappers/generated/operator_wrapper"
	"github.com/smartcontractkit/chainlink/core/logger"
	"github.com/smartcontractkit/chainlink/core/services/eth"
	"github.com/smartcontractkit/chainlink/core/services/job"
	"github.com/smartcontractkit/chainlink/core/services/log"
	"github.com/smartcontractkit/chainlink/core/services/pipeline"
	"github.com/smartcontractkit/chainlink/core/store/models"
	"github.com/smartcontractkit/chainlink/core/utils"
	"gorm.io/gorm"
)

type (
	Delegate struct {
		logBroadcaster log.Broadcaster
		pipelineRunner pipeline.Runner
		pipelineORM    pipeline.ORM
		db             *gorm.DB
		ethClient      eth.Client
		chHeads        chan models.Head
		config         Config
	}

	Config interface {
		MinIncomingConfirmations() uint32
		MinimumContractPayment() *assets.Link
	}
)

var _ job.Delegate = (*Delegate)(nil)

func NewDelegate(
	logBroadcaster log.Broadcaster,
	pipelineRunner pipeline.Runner,
	pipelineORM pipeline.ORM,
	ethClient eth.Client,
	db *gorm.DB,
	config Config,
) *Delegate {
	return &Delegate{
		logBroadcaster,
		pipelineRunner,
		pipelineORM,
		db,
		ethClient,
		make(chan models.Head, 1),
		config,
	}
}

func (d *Delegate) JobType() job.Type {
	return job.DirectRequest
}

func (Delegate) AfterJobCreated(spec job.Job)  {}
func (Delegate) BeforeJobDeleted(spec job.Job) {}

// ServicesForSpec returns the log listener service for a direct request job
func (d *Delegate) ServicesForSpec(job job.Job) (services []job.Service, err error) {
	if job.DirectRequestSpec == nil {
		return nil, errors.Errorf("directrequest.Delegate expects a *job.DirectRequestSpec to be present, got %v", job)
	}
	concreteSpec := job.DirectRequestSpec

	oracle, err := operator_wrapper.NewOperator(concreteSpec.ContractAddress.Address(), d.ethClient)
	if err != nil {
		return
	}

	minIncomingConfirmations := d.config.MinIncomingConfirmations()

	if concreteSpec.MinIncomingConfirmations.Uint32 > minIncomingConfirmations {
		minIncomingConfirmations = concreteSpec.MinIncomingConfirmations.Uint32
	}

	logListener := &listener{
		config:                   d.config,
		logBroadcaster:           d.logBroadcaster,
		oracle:                   oracle,
		pipelineRunner:           d.pipelineRunner,
		db:                       d.db,
		pipelineORM:              d.pipelineORM,
		job:                      job,
		mbLogs:                   utils.NewMailbox(50),
		minIncomingConfirmations: uint64(minIncomingConfirmations),
		requesters:               concreteSpec.Requesters,
		minContractPayment:       concreteSpec.MinContractPayment,
		chStop:                   make(chan struct{}),
	}
	services = append(services, logListener)

	return
}

var (
	_ log.Listener = &listener{}
	_ job.Service  = &listener{}
)

type listener struct {
	config                   Config
	logBroadcaster           log.Broadcaster
	oracle                   operator_wrapper.OperatorInterface
	pipelineRunner           pipeline.Runner
	db                       *gorm.DB
	pipelineORM              pipeline.ORM
	job                      job.Job
	runs                     sync.Map
	shutdownWaitGroup        sync.WaitGroup
	mbLogs                   *utils.Mailbox
	minIncomingConfirmations uint64
	requesters               models.AddressCollection
	minContractPayment       *assets.Link
	chStop                   chan struct{}
	utils.StartStopOnce
}

// Start complies with job.Service
func (l *listener) Start() error {
	return l.StartOnce("DirectRequestListener", func() error {
		unsubscribeLogs := l.logBroadcaster.Register(l, log.ListenerOpts{
			Contract: l.oracle.Address(),
			ParseLog: l.oracle.ParseLog,
			LogsWithTopics: map[common.Hash][][]log.Topic{
				operator_wrapper.OperatorOracleRequest{}.Topic():       {{log.Topic(l.job.ExternalIDEncodeBytesToTopic()), log.Topic(l.job.ExternalIDEncodeStringToTopic())}},
				operator_wrapper.OperatorCancelOracleRequest{}.Topic(): {{log.Topic(l.job.ExternalIDEncodeBytesToTopic()), log.Topic(l.job.ExternalIDEncodeStringToTopic())}},
			},
			NumConfirmations: l.minIncomingConfirmations,
		})
		l.shutdownWaitGroup.Add(2)
		go l.run()

		go func() {
			<-l.chStop
			unsubscribeLogs()
			l.shutdownWaitGroup.Done()
		}()

		return nil
	})
}

// Close complies with job.Service
func (l *listener) Close() error {
	return l.StopOnce("DirectRequestListener", func() error {
		l.runs.Range(func(key, runCloserChannelIf interface{}) bool {
			runCloserChannel, _ := runCloserChannelIf.(chan struct{})
			close(runCloserChannel)
			return true
		})
		l.runs = sync.Map{}

		close(l.chStop)
		l.shutdownWaitGroup.Wait()

		return nil
	})
}

func (l *listener) HandleLog(lb log.Broadcast) {
	wasOverCapacity := l.mbLogs.Deliver(lb)
	if wasOverCapacity {
		logger.Error("DirectRequestListener: log mailbox is over capacity - dropped the oldest log")
	}
}

func (l *listener) run() {
	for {
		select {
		case <-l.chStop:
			l.shutdownWaitGroup.Done()
			return
		case <-l.mbLogs.Notify():
			l.handleReceivedLogs()
		}
	}
}

func (l *listener) handleReceivedLogs() {
	for {
		i, exists := l.mbLogs.Retrieve()
		if !exists {
			return
		}
		lb, ok := i.(log.Broadcast)
		if !ok {
			panic(errors.Errorf("DirectRequestListener: invariant violation, expected log.Broadcast but got %T", lb))
		}
		ctx, cancel := postgres.DefaultQueryCtx()
		was, err := l.logBroadcaster.WasAlreadyConsumed(l.db.WithContext(ctx), lb)
		cancel()
		if err != nil {
			logger.Errorw("DirectRequestListener: could not determine if log was already consumed", "error", err)
			return
		} else if was {
			return
		}

		logJobSpecID := lb.RawLog().Topics[1]
		if logJobSpecID == (common.Hash{}) || (logJobSpecID != l.job.ExternalIDEncodeStringToTopic() && logJobSpecID != l.job.ExternalIDEncodeBytesToTopic()) {
			logger.Debugw("DirectRequestListener: Skipping Run for Log with wrong Job ID", "logJobSpecID", logJobSpecID, "actualJobID", l.job.ExternalJobID)
			return
		}

		log := lb.DecodedLog()
		if log == nil || reflect.ValueOf(log).IsNil() {
			logger.Error("DirectRequestListener: HandleLog: ignoring nil value")
			return
		}

		switch log := log.(type) {
		case *operator_wrapper.OperatorOracleRequest:
			l.handleOracleRequest(log, lb)
		case *operator_wrapper.OperatorCancelOracleRequest:
			l.handleCancelOracleRequest(log, lb)
		default:
			logger.Warnf("unexpected log type %T", log)
		}
	}
}

func oracleRequestToMap(request *operator_wrapper.OperatorOracleRequest) map[string]interface{} {
	result := make(map[string]interface{})
	result["specId"] = fmt.Sprintf("0x%x", request.SpecId)
	result["requester"] = request.Requester.Hex()
	result["requestId"] = formatRequestId(request.RequestId)
	result["payment"] = fmt.Sprintf("%v", request.Payment)
	result["callbackAddr"] = request.CallbackAddr.Hex()
	result["callbackFunctionId"] = fmt.Sprintf("0x%x", request.CallbackFunctionId)
	result["cancelExpiration"] = fmt.Sprintf("%v", request.CancelExpiration)
	result["dataVersion"] = fmt.Sprintf("%v", request.DataVersion)
	result["data"] = fmt.Sprintf("0x%x", request.Data)
	return result
}

func (l *listener) handleOracleRequest(request *operator_wrapper.OperatorOracleRequest, lb log.Broadcast) {
	logger.Infow("Direct request: oracle request received",
		"specId", fmt.Sprintf("%0x", request.SpecId),
		"requester", request.Requester,
		"requestId", fmt.Sprintf("%0x", request.RequestId),
		"payment", request.Payment,
		"callbackAddr", request.CallbackAddr,
		"callbackFunctionId", fmt.Sprintf("%0x", request.CallbackFunctionId),
		"cancelExpiration", request.CancelExpiration,
		"dataVersion", request.DataVersion,
		"data", fmt.Sprintf("%0x", request.Data),
	)

	if !l.allowRequester(request.Requester) {
		logger.Infow("Rejected run for invalid requester",
			"requester", request.Requester,
			"allowedRequesters", l.requesters.ToStrings(),
		)
		ctx, cancel := postgres.DefaultQueryCtx()
		defer cancel()
		if err := l.logBroadcaster.MarkConsumed(l.db.WithContext(ctx), lb); err != nil {
			logger.Errorw("DirectRequest: unable to mark log consumed", "err", err, "log", lb.String())
		}
		return
	}

	var minContractPayment *assets.Link
	if l.minContractPayment != nil {
		minContractPayment = l.minContractPayment
	} else {
		minContractPayment = l.config.MinimumContractPayment()
	}
	if minContractPayment != nil {
		requestPayment := assets.Link(*request.Payment)
		if minContractPayment.Cmp(&requestPayment) > 0 {
			logger.Infow("Rejected run for insufficient payment",
				"minContractPayment", minContractPayment.String(),
				"requestPayment", requestPayment.String(),
			)
			ctx, cancel := postgres.DefaultQueryCtx()
			defer cancel()
			if err := l.logBroadcaster.MarkConsumed(l.db.WithContext(ctx), lb); err != nil {
				logger.Errorw("DirectRequest: unable to mark log consumed", "err", err, "log", lb.String())
			}
			return
		}
	}

	meta := make(map[string]interface{})
	meta["oracleRequest"] = oracleRequestToMap(request)

	logger := logger.CreateLogger(logger.Default.With(
		"jobName", l.job.PipelineSpec.JobName,
		"jobID", l.job.PipelineSpec.JobID,
	))

	l.shutdownWaitGroup.Add(1)
	go func() {
		defer l.shutdownWaitGroup.Done()

		runCloserChannel := make(chan struct{})
		runCloserChannelIf, loaded := l.runs.LoadOrStore(formatRequestId(request.RequestId), runCloserChannel)
		if loaded {
			runCloserChannel, _ = runCloserChannelIf.(chan struct{})
		}
		ctx, cancel := utils.CombinedContext(runCloserChannel, context.Background())
		defer cancel()

		vars := pipeline.NewVarsFrom(map[string]interface{}{
			"jobSpec": map[string]interface{}{
				"databaseID":    l.job.ID,
				"externalJobID": l.job.ExternalJobID,
				"name":          l.job.Name.ValueOrZero(),
			},
			"jobRun": map[string]interface{}{
				"meta":           meta,
				"logBlockHash":   request.Raw.BlockHash,
				"logBlockNumber": request.Raw.BlockNumber,
				"logTxHash":      request.Raw.TxHash,
				"logAddress":     request.Raw.Address,
				"logTopics":      request.Raw.Topics,
				"logData":        request.Raw.Data,
			},
		})

		run, trrs, err := l.pipelineRunner.ExecuteRun(ctx, *l.job.PipelineSpec, vars, *logger)
		if ctx.Err() != nil {
			return
		} else if err != nil {
			logger.Errorw("DirectRequest failed to create run", "err", err)
		}
		ctx, cancel = context.WithTimeout(ctx, postgres.DefaultQueryTimeout)
		defer cancel()
		err = postgres.GormTransaction(ctx, l.db, func(tx *gorm.DB) error {
			_, err = l.pipelineRunner.InsertFinishedRun(tx, run, trrs, true)
			if err != nil {
				return err
			}
			return l.logBroadcaster.MarkConsumed(tx, lb)
		})
		if ctx.Err() != nil {
			return
		} else if err != nil {
			logger.Errorw("DirectRequest failed to create run", "err", err)
		}
	}()
}

func (l *listener) allowRequester(requester common.Address) bool {
	if len(l.requesters) == 0 {
		return true
	}
	for _, addr := range l.requesters {
		if addr == requester {
			return true
		}
	}
	return false
}

// Cancels runs that haven't been started yet, with the given request ID
func (l *listener) handleCancelOracleRequest(request *operator_wrapper.OperatorCancelOracleRequest, lb log.Broadcast) {
	runCloserChannelIf, loaded := l.runs.LoadAndDelete(formatRequestId(request.RequestId))
	if loaded {
		close(runCloserChannelIf.(chan struct{}))
	}
	ctx, cancel := postgres.DefaultQueryCtx()
	defer cancel()
	if err := l.logBroadcaster.MarkConsumed(l.db.WithContext(ctx), lb); err != nil {
		logger.Errorw("DirectRequest: failed to mark log consumed", "log", lb.String())
	}
}

// JobID complies with log.Listener
func (*listener) JobID() models.JobID {
	return models.NilJobID
}

// Job complies with log.Listener
func (l *listener) JobIDV2() int32 {
	return l.job.ID
}

// IsV2Job complies with log.Listener
func (*listener) IsV2Job() bool {
	return true
}

func formatRequestId(requestId [32]byte) string {
	return fmt.Sprintf("0x%x", requestId)
}
